{ Version 010921. Copyright  Alexey A.Chernobaev, 1996-2001 }

unit ParseCmd;
{
  Command line parsing.
    .

  Example:
   :

  CmdParser:=TCommandLineParser.Create([], False);
  try
    CmdParser.Parse(Windows.GetCommandLine, 1);

    for I:=0 to CmdParser.ParamCount - 1 do
      if CmdParser.ParamStr(I) = ...

    for I:=0 to CmdParser.SimpleParamCount - 1 do
      if CmdParser.SimpleParamStr(I) = ...

    for I:=0 to CmdParser.OptionCount - 1 do
      if CmdParser.OptionStr(I) = ...

    if CmdParser.HasOption('from') then ...
  finally
    CmdParser.Free;
  end;
}

interface

{$I VCheck.inc}

uses
  SysUtils, Aliasv, Int16v, StrLst;

type
  TCharSet = set of Char;

  TCommandLineParser = class
  protected
    FOptionPrefix: TCharSet;
    FParamsList, FSimpleList, FOptionKeys, FOptionValues: TStrLst;
    FOptionIndex: TIntegerVector;
  public
    constructor Create(OptionPrefix: TCharSet; CaseSensitive: Boolean);
    { OptionPrefix:    ;  [],  
         ['/', '-'];  CaseSensitive = True,  ,
         ,  ,   
       }
    { OptionPrefix: command line option prefixes; if [] then default values
      ['/', '-'] are used; if CaseSensitive = True then options in different
      case are interpreted as different, otherwise as equivalent. }
    destructor Destroy; override;
    procedure Parse(const CommandLine: String; FromIndex: Integer;
      QuotesDoubling: Boolean{$IFDEF V_D4} = False{$ENDIF});
    {    CommandLine;    
       FromIndex ;  QuotesDoubling = True,  
          '"',     }
    { parses the string CommandLine ignoring first FromIndex parameters; if
      QuotesDoubling = True then the character '"' is allowed among the
      elements of the string }
    function ParamCount: Integer;
    {    }
    { total number of parameters }
    function ParamStr(I: Integer): String;
    { I-  (I = 0..ParamCount - 1);  I >= ParamCount, 
         }
    { Ith parameter (I = 0..ParamCount - 1); if I >= ParamCount then returns
      empty string }
    function SimpleParamCount: Integer;
    {    (..    OptionPrefix) }
    { number of simple parameters (i.e. ones not beginning with OptionPrefix) }
    function SimpleParamStr(I: Integer): String;
    { I-   (I = 0..SimpleParamCount - 1); 
      I >= SimpleParamCount,     }
    { Ith simple parameter (I = 0..SimpleParamCount - 1); if
      I >= SimpleParamCount then returns empty string }
    function OptionCount: Integer;
    {   (.. ,   OptionPrefix) }
    { number of options (i.e. parameters beginning with OptionPrefix) }
    function OptionStr(I: Integer): String;
    { I-  (I = 0..OptionCount - 1) ;  I >= OptionCount,
          }
    { Ith option (I = 0..OptionCount - 1); if I >= OptionCount then empty string
      will be returned }
    function HasOption(const OptionName: String): Boolean;
    {  True,       OptionName;
      ,  Parse('/option')  HasOption('option') 
      True ( '/'   AnOptionPrefix) }
    { returns True if command line contains options OptionName; e.g. after
      Parse('/option') call to HasOption('option') will return True (if '/' is
      in AnOptionPrefix) }
    function OptionValue(const OptionName: String): String;
    {    OptionName; , 
      Parse('/option:value')  OptionValue('option')  'value';
       HasOption(OptionName) = False,     }
    { returns value of option OptionName; e.g. after Parse('/option:value')
      call to OptionValue('option') will return 'value'; if
      HasOption(OptionName) = False then empty string will be returned }
  end;

implementation

{$IFDEF CHECK_OBJECTS_FREE}
uses ChckFree;
{$ENDIF}

constructor TCommandLineParser.Create(OptionPrefix: TCharSet; CaseSensitive: Boolean);
begin
  inherited Create;
  FParamsList:=TStrLst.Create;
  FSimpleList:=TStrLst.Create;
  if CaseSensitive then
    FOptionKeys:=TCaseSensStrLst.Create
  else
    FOptionKeys:=TStrLst.Create;
  FOptionValues:=TStrLst.Create;
  FOptionIndex:=TIntegerVector.Create(0, 0);
  if OptionPrefix <> [] then
    FOptionPrefix:=OptionPrefix
  else
    FOptionPrefix:=['/', '-'];
  {$IFDEF CHECK_OBJECTS_FREE}
  RegisterObjectCreate(Self);
  {$ENDIF}
end;

destructor TCommandLineParser.Destroy;
begin
  {$IFDEF CHECK_OBJECTS_FREE}
  RegisterObjectFree(Self);
  {$ENDIF}
  FParamsList.Free;
  FSimpleList.Free;
  FOptionKeys.Free;
  FOptionValues.Free;
  FOptionIndex.Free;
  inherited Destroy;
end;

procedure TCommandLineParser.Parse(const CommandLine: String; FromIndex: Integer;
  QuotesDoubling: Boolean{$IFDEF V_D4} = False{$ENDIF});
var
  ParamIndex: Integer;
  S: String;

  procedure ProcessParam;
  var
    I, J: Integer;
    S1, S2: String;
  begin
    if S <> '' then begin
      Inc(ParamIndex);
      if ParamIndex > FromIndex then begin
        I:=FParamsList.Add(S);
        if S[1] in FOptionPrefix then begin
          FOptionIndex.Add(I);
          I:=Pos(':', S);
          if I > 0 then begin
            S1:=Copy(S, 2, I - 2);
            S2:=Copy(S, I + 1, Length(S));
            J:=FOptionKeys.IndexOf(S1);
            if J < 0 then begin
              FOptionKeys.Add(S1);
              FOptionValues.Add(S2);
            end
            else
              FOptionValues[J]:=S2;
          end
          else begin
            FOptionKeys.Add(Copy(S, 2, Length(S)));
            FOptionValues.Add('');
          end;
        end
        else
          FSimpleList.Add(S);
      end;
      S:='';
    end;
  end;

var
  I, L: Integer;
  C: Char;
  Quote: Boolean;
begin
  FParamsList.Clear;
  FSimpleList.Clear;
  FOptionKeys.Clear;
  FOptionValues.Clear;
  FOptionIndex.Clear;
  ParamIndex:=0;
  S:='';
  Quote:=False;
  L:=Length(CommandLine);
  I:=1;
  while I <= L do begin
    C:=CommandLine[I];
    Case C of
      '"':
        begin
          S:=S + C;
          if Quote then begin
            if QuotesDoubling and (I < L) and (CommandLine[I + 1] = '"') then begin
              S:=S + '"';
              Inc(I, 2);
              Continue;
            end;
            ProcessParam;
          end;
          Quote:=not Quote;
        end;
      ' ':
        if Quote then S:=S + C else ProcessParam;
    Else
      S:=S + C;
    End;
    Inc(I);
  end;
  ProcessParam;
  FOptionKeys.SortWith(FOptionValues);
end;

function TCommandLineParser.ParamCount: Integer;
begin
  Result:=FParamsList.Count;
end;

function TCommandLineParser.ParamStr(I: Integer): String;
begin
  if (I >= 0) and (I < FParamsList.Count) then
    Result:=FParamsList.Items[I]
  else
    Result:='';
end;

function TCommandLineParser.SimpleParamCount: Integer;
begin
  Result:=FSimpleList.Count;
end;

function TCommandLineParser.SimpleParamStr(I: Integer): String;
begin
  if (I >= 0) and (I < FSimpleList.Count) then
    Result:=FSimpleList.Items[I]
  else
    Result:='';
end;

function TCommandLineParser.OptionCount: Integer;
begin
  Result:=FOptionKeys.Count;
end;

function TCommandLineParser.OptionStr(I: Integer): String;
begin
  if (I >= 0) and (I < FOptionIndex.Count) then
    Result:=FParamsList.Items[FOptionIndex[I]]
  else
    Result:='';
end;

function TCommandLineParser.HasOption(const OptionName: String): Boolean;
begin
  Result:=FOptionKeys.FindInSorted(OptionName) >= 0;
end;

function TCommandLineParser.OptionValue(const OptionName: String): String;
var
  I: Integer;
begin
  I:=FOptionKeys.FindInSorted(OptionName);
  if I >= 0 then
    Result:=FOptionValues[I]
  else
    Result:='';
end;

end.
