{ Version 010712. Copyright  Alexey A.Chernobaev, 1996-2001 }

unit VMapStrm;
{
  Memory-mapped streams (Win32). This code is partly based on the
  TMemoryMappedFileStream class by Dmitry Streblechenko.
}

interface

{$I VCheck.inc}

uses
  Windows, SysUtils, ExtType, ExtSys, VStream, VFStream, VectErr;

type
  TVMemoryMappedStream = class(TVFileStream)
  protected
    FMemory: Pointer;
    FMapHandle: THandle;
    FAccess, { file access attributes }
    FShare, { file sharing attributes }
    FPageProtect, { mapping object page protection attributes }
    FMemProtect: DWORD; { protection of the mapped memory }
    FCapacity: Int32;
    FMapName: String;
    FNewMapping: Boolean;
    procedure SetAttributes(Mode: Word);
    procedure CreateMapping;
    procedure CloseHandles;
    function GetSize: Int32; override;
    procedure SetSize(NewSize: Int32); override;
    procedure SetCapacity(NewCapacity:  Int32);
  public
    constructor Create(const AFileName, AMapName: String; Mode: Word);
    constructor CreateWithCapacity(const AFileName, AMapName: String; Mode: Word;
      ACapacity: Int32);
    constructor Open(const AMapName: String; Mode: Word);
    destructor Destroy; override;
    procedure Seek(Offset: Int32); override;
    procedure WriteProc(const Buffer; Count: Int32); override;
    function ReadFunc(var Buffer; Count: Int32): Int32; override;
    function Flush: Boolean; override;
    function Search(const SearchBytes; SearchCount: Integer;
      MaxBytesRead: ILong): Bool; override;
    property Memory: Pointer read FMemory;
    property Capacity: Int32 read FCapacity;
    property MapName: String read FMapName;
    { name of the file mapping object }
    property NewMapping: Boolean read FNewMapping;
    { allows to determine whether the file mapping object existed before
      creation of Self }
  end;

implementation

{ TVMemoryMappedStream }

const
  MemoryDelta = 65536; { file size increment; must be degree of 2! }

procedure TVMemoryMappedStream.SetAttributes(Mode: Word);
begin
  FAccess:=DWORD(GENERIC_READ or GENERIC_WRITE);
  FPageProtect:=PAGE_READWRITE or SEC_COMMIT or SEC_NOCACHE;
  if Mode and fmOpenReadWrite <> 0 then
    FMemProtect:=FILE_MAP_ALL_ACCESS
  else if Mode and fmOpenWrite <> 0 then
    FMemProtect:=FILE_MAP_WRITE
  else begin { all others including fmOpenRead = 0 }
    FAccess:=DWORD(GENERIC_READ);
    FPageProtect:=PAGE_READONLY or SEC_COMMIT or SEC_NOCACHE;
    FMemProtect:=FILE_MAP_READ;
  end;
  if Mode and fmShareDenyWrite = fmShareDenyWrite then FShare:=FILE_SHARE_READ;
  if Mode and fmShareDenyRead = fmShareDenyRead then FShare:=FILE_SHARE_WRITE;
  if Mode and fmShareExclusive = fmShareExclusive then FShare:=0;
end;

procedure TVMemoryMappedStream.CreateMapping;
begin
  FMapHandle:=CreateFileMapping(THandle(FHandle), nil, FPageProtect, 0,
    FCapacity, PChar(FMapName));
  if FMapHandle = 0 then
    raise EVStream.CreateFmt(SCreateMappingError, [FMapName, GetLastError]);
  FNewMapping:=GetLastError = 0;
  FMemory:=MapViewOfFile(FMapHandle, FMemProtect, 0, 0, 0);
  if FMemory = nil then
    raise EVStream.CreateFmt(SFileMappingError, [FFileName, GetLastError]);
end;

constructor TVMemoryMappedStream.Create(const AFileName, AMapName: String; Mode: Word);
begin
  CreateWithCapacity(AFileName, AMapName, Mode, MemoryDelta);
end;

constructor TVMemoryMappedStream.CreateWithCapacity(const AFileName, AMapName: String;
  Mode: Word; ACapacity: Int32);
var
  Err: Int32;
begin
  CreateInherited;
  FFileName:=AFileName;
  FMapName:=AMapName;
  SetAttributes(Mode);
  FCapacity:=ACapacity;
  if AFileName <> '' then
    if Mode = fmCreate then begin
      FHandle:=CreateFile(PChar(AFileName), FAccess, FShare, nil, CREATE_ALWAYS, 0, 0);
      if THandle(FHandle) = INVALID_HANDLE_VALUE then Error(SFileCreateError);
    end
    else begin
      FHandle:=CreateFile(PChar(AFileName), FAccess, FShare, nil, OPEN_EXISTING, 0, 0);
      if THandle(FHandle) = INVALID_HANDLE_VALUE then Error(SFileCreateError);
      FSize:=GetFileSize(FHandle, @Err);
      if (Err <> 0) or (FSize < 0) then Error(SFileTooLong);
      if (FSize > ACapacity) and (FAccess and GENERIC_WRITE <> 0) then
        FCapacity:=FSize;
    end
  else begin
    DWORD(FHandle):=DWORD($FFFFFFFF);
    FSize:=ACapacity;
  end;
  CreateMapping;
end;

constructor TVMemoryMappedStream.Open(const AMapName: String; Mode: Word);
begin
  CreateInherited;
  FMapName:=AMapName;
  SetAttributes(Mode);
  FSize:=MemoryDelta;
  FCapacity:=MemoryDelta;
  THandle(FHandle):=INVALID_HANDLE_VALUE;
  FMapHandle:=OpenFileMapping(FMemProtect, False, PChar(AMapName));
  if FMapHandle = 0 then
    raise EVStream.CreateFmt(SOpenMappingError, [AMapName, GetLastError]);
  FMemory:=MapViewOfFile(FMapHandle, FMemProtect, 0, 0, 0);
  if FMemory = nil then
    raise EVStream.CreateFmt(SFileMappingError, [AMapName, GetLastError]);
end;

procedure TVMemoryMappedStream.CloseHandles;
begin
  try
    if FMemory <> nil then
      if not (FlushViewOfFile(FMemory, 0) and UnmapViewOfFile(FMemory)) then
        Error(SFlushError);
  finally
    FMemory:=nil;
    if FMapHandle <> 0 then Windows.CloseHandle(FMapHandle);
  end;
end;

destructor TVMemoryMappedStream.Destroy;
begin
  try
    CloseHandles;
  finally
    if (FAccess <> GENERIC_READ) and (FileSeek(FHandle, FSize, 0) = FSize) then
      SetEndOfFile(FHandle);
    inherited Destroy;
  end;
end;

function TVMemoryMappedStream.GetSize: Int32;
begin
  Result:=FSize;
end;

procedure TVMemoryMappedStream.SetSize(NewSize: Int32);
begin
  if NewSize <> FSize then begin
    SetCapacity(NewSize);
    FSize:=NewSize;
    if FPosition > NewSize then FPosition:=NewSize;
  end;
end;

procedure TVMemoryMappedStream.SetCapacity(NewCapacity: Int32);
var
  Err: Int32;
begin
  NewCapacity:=(NewCapacity + MemoryDelta - 1) and -MemoryDelta;
  if FCapacity <> NewCapacity then begin
    { Close MMF and try to recreate it with the new size.
      FHandle = INVALID_HANDLE_VALUE for the opened (not created!) named mapping object. }
    if THandle(FHandle) <> INVALID_HANDLE_VALUE then begin
      CloseHandles;
      CloseHandle;
      FHandle:=CreateFile(PChar(FFileName), FAccess, FShare, nil, OPEN_EXISTING, 0, 0);
      if THandle(FHandle) = INVALID_HANDLE_VALUE then Error(SFileCreateError);
      FSize:=GetFileSize(FHandle, @Err);
      if (Err <> 0) or (FSize < 0) then Error(SFileTooLong);
    end
    else
      Error(SSetSizeError);
    FCapacity:=NewCapacity;
    CreateMapping;
  end;
end;

procedure TVMemoryMappedStream.Seek(Offset: Int32);
begin
  if (Offset < 0) or (Offset > FSize) then Error(Format(SSeekError, [Offset]));
  FPosition:=Offset;
end;

procedure TVMemoryMappedStream.WriteProc(const Buffer; Count: Int32);
var
  Pos: Int32;
begin
  if FSize < 0 then Error(SFileWriteError); { FSize = -1   CloseHandle }
  if Count >= 0 then begin
    Pos:=FPosition + Count;
    if Pos > 0 then begin
      if Pos > FSize then begin
        if Pos > FCapacity then SetCapacity(Pos);
        FSize:=Pos;
      end;
      Move(Buffer, (PChar(FMemory) + FPosition)^, Count);
      FPosition:=Pos;
    end;
  end;
end;

function TVMemoryMappedStream.ReadFunc(var Buffer; Count: Int32): Int32;
begin
  if Count > 0 then begin
    Result:=FSize - FPosition;
    if Result < 0 then Error(SFileReadError);
    if Result > Count then Result:=Count;
    Move((PChar(FMemory) + FPosition)^, Buffer, Result);
    Inc(FPosition, Result);
  end
  else
    Result:=0;
end;

function TVMemoryMappedStream.Flush: Boolean;
begin
  Result:=FlushViewOfFile(FMemory, 0);
end;

function TVMemoryMappedStream.Search(const SearchBytes; SearchCount: Integer;
  MaxBytesRead: ILong): Bool;
var
  I: Integer;
begin
  I:=FSize - FPosition;
  if I < 0 then Error(SFileReadError);
  if MaxBytesRead > I then MaxBytesRead:=I;
  I:=FindInBuf(SearchBytes, SearchCount, (PChar(FMemory) + FPosition)^, MaxBytesRead);
  if I >= 0 then begin
    Inc(FPosition, I);
    Result:=True;
  end
  else
    Result:=False;
end;

end.
