{ Version 001216. Copyright  Alexey A.Chernobaev, 1996-2000 }

unit Vectors;

interface

{$I VCheck.inc}

uses
  SysUtils, ExtSys, ExtType, VStream, VectErr;

const
  MaxElemSize = 1024;
  {     (. !!! ) }
  { maximum vector element size (see !!! marks) }

type
  TElemSize = 1..MaxElemSize;

  EVectorError = class(Exception);

  TVectorObject = class
    class procedure Error(const Msg: String; const Data: array of const);
  end;

  {    }
  { base vector class }
  TVector = class(TVectorObject)
  protected
    FItems: PRegular;
    FCount, FCapacity: Integer;
    FElemSize: TElemSize;
    function GetCount: Integer; virtual;
    procedure SetCount(ACount: Integer); virtual;
    function ItemsSize: Integer;
    procedure ChangeSize(ACount: Integer);
    procedure Resize(ACount: Integer);
    procedure InitMemory(Offset, InitCount: Integer); virtual;
    procedure Expand(I: Integer); virtual;
    {          
       >= I     }
    { increases vector size by one and moves all elements with indexes >= I up }
    procedure SetCapacity(NewCapacity: Integer); virtual;
    procedure RawAssign(Source: TVector);
    procedure RawCopyRange(FromIndex: Integer; Source: TVector; SourceFrom,
      SourceTo: Integer);
    procedure RawExchangeRange(FromIndex, ToIndex, ACount: Integer);
    function RawEqualTo(V: TVector): Bool;
  public
    constructor Create(AnElemSize: TElemSize);
    {   (Count = 0)     ElemSize  }
    { creates empty (Count = 0) vector with elements of size ElemSize bytes }
    destructor Destroy; override;
    procedure WriteToStream(VStream: TVStream); virtual;
    {     }
    { writes vector to the stream }
    procedure ReadFromStream(VStream: TVStream); virtual;
    {     }
    { reads vector from the stream }
    procedure Assign(Source: TVector); virtual;
    {      Source }
    { assigns values of vector Source to the current vector }
    procedure CopyRange(FromIndex: Integer; Source: TVector; SourceFrom,
      SourceTo: Integer); virtual;
    {      [SourceFrom..SourceTo]  Source,
          ,   FromIndex;  
          }
    { copies elements [SourceFrom..SourceTo] of the vector Source to the current
      vector, placing them from the index FromIndex; increases the length of the
      current vector if needed }
    procedure ExchangeRange(FromIndex, ToIndex, ACount: Integer); virtual;
    {   ACount :   FromIndex  ToIndex,
      FromIndex+1  ToIndex+1, ... FromIndex+ACount-1  ToIndex+ACount-1;
           }
    { exchanges ACount elements: with indexes FromIndex and ToIndex,
      FromIndex+1 and ToIndex+1, ... FromIndex+ACount-1 and ToIndex+ACount-1;
      element ranges must not overlap }
    function EqualTo(V: TVector): Bool; virtual;
    {  ;   ,    (FCount) 
       (FItems^)  }
    { compares vectors; vectors are considered equal if both their sizes
      (FCount) and data (FItem^) are equal }
    procedure Delete(I: Integer); virtual;
    {  I-    }
    { deletes element I from the vector }
    procedure DeleteRange(I, ACount: Integer); virtual;
    {  ACount > 0,   ACount ,   I (:
      I + ACount <= Count),     }
    { if ACount > 0 then deletes ACount elements beginning from I (precondition:
      I + ACount <= Count), otherwise does nothing }
    procedure Clear; virtual;
    {   (Count:=0) }
    { clears the vector (Count:=0) }
    function SizeInBytes: Integer; virtual;
    {     ,    ,    ,
              (.
      Capacity, Quantum) }
    { returns vector size in bytes including all fields but not considering the
      fact that actually memory is allocated by blocks (see Capacity, Quantum) }
    procedure Pack; virtual;
    { Capacity:=FCount }
    procedure Grow(Delta: Integer);
    { Count:=Count + Delta }
    property ElemSize: TElemSize read FElemSize;
    {    }
    { vector element size }
    property Count: Integer read GetCount write SetCount;
    {   }
    { number of elements in the vector }
    property Capacity: Integer read FCapacity write SetCapacity;
    {         ElemSize   }
    { actual vector size in blocks by ElemSize bytes each }
    property Memory: PRegular read FItems;
    {  }
    { raw data }
  end;

  TSortableVector = class(TVector)
  public
    function Compare(I: Integer; const V): Int32; virtual; abstract;
    {    ,  I-   V, , 
       ,    ,  I-   V }
    { returns number greater then zero if element I is greater then V, zero
      if they are equal and number less then zero if element I is less then V }
    procedure Exchange(I, J: Integer); virtual; abstract;
    {      I  J }
    { exchanges elements with indexes I and J }
    procedure GetUntyped(I: Integer; var Result); virtual; abstract;
    {  I-  ( Result     ElemSize) }
    { returns element I (size of Result can't be less then ElemSize) }
    procedure DoSortRange(L, R: Integer); virtual;
    {    L  R       }
    { sorts elements from L to R ascending without range checking }
    procedure DoSortRangeDesc(L, R: Integer); virtual;
    {    L  R       }
    { sorts elements from L to R descending without range checking }
    procedure SortRange(L, R: Integer);
    {    L  R   (L, R = 0..Count - 1) }
    { sorts elements from L to R ascending (L, R = 0..Count - 1) }
    procedure SortRangeDesc(L, R: Integer);
    {    L  R   (L, R = 0..Count - 1) }
    { sorts elements from L to R descending (L, R = 0..Count - 1) }
    procedure Sort; virtual;
    {     }
    { sorts vector ascending }
    procedure SortDesc; virtual;
    {     }
    { sorts vector descending }
    procedure DoSortRangeWith(L, R: Integer; AVector: TSortableVector);
    procedure DoSortRangeDescWith(L, R: Integer; AVector: TSortableVector);
    procedure SortRangeWith(L, R: Integer; AVector: TSortableVector);
    procedure SortRangeDescWith(L, R: Integer; AVector: TSortableVector);
    procedure SortWith(AVector: TSortableVector);
    {    ,    
          AVector; AVector   ,  ,
          }
    { sorts vector ascending preserving correspondence between elements of the
      current vector and vector AVector; size of AVector can not be less then
      the size of the current vector }
    procedure SortDescWith(AVector: TSortableVector);
    {    ,    
          AVector; AVector   ,  ,
          }
    { sorts vector descending preserving correspondence between elements of the
      current vector and vector AVector; size of AVector can not be less then
      size of the current vector }
    procedure SortWithArray(VectorArray: array of TSortableVector);
    {  SortWith     }
    { analog of SortWith with variable number of parameters }
    procedure SortDescWithArray(VectorArray: array of TSortableVector);
    {  SortWithDesc     }
    { analog of SortWithDesc with variable number of parameters }
    procedure Invert;
    {   (   ,  -
      ,  ..) }
    { inverts the vector (last element becomes the first, next to last - the
      second, etc.) }
  end;

  TVectorClass = class of TVector;
  TSortableVectorClass = class of TSortableVector;

implementation

{$IFDEF CHECK_OBJECTS_FREE}
uses ChckFree;
{$ENDIF}

{ TVectorObject }

class procedure TVectorObject.Error(const Msg: String; const Data: array of const);
{$IFDEF V_DELPHI}{$IFDEF WIN32}
  function ReturnAddr: Pointer;
  asm
          mov     eax, [ebp+4]
  end;
{$ENDIF}{$ENDIF}
begin
  raise EVectorError.Create('Class ' + ClassName + ': ' + ErrMsg(Msg, Data))
    {$IFDEF V_DELPHI}{$IFDEF WIN32}at ReturnAddr{$ENDIF}{$ENDIF};
end;

{ TVector }

constructor TVector.Create(AnElemSize: TElemSize);
begin
  {$IFDEF CHECK_OBJECTS_FREE}
  RegisterObjectCreate(Self);
  {$ENDIF}
  inherited Create;
  FElemSize:=AnElemSize;
end;

destructor TVector.Destroy;
begin
  {$IFDEF CHECK_OBJECTS_FREE}
  RegisterObjectFree(Self);
  {$ENDIF}
  ChangeSize(0);
  inherited Destroy;
end;

function TVector.GetCount: Integer;
begin
  Result:=FCount;
end;

procedure TVector.InitMemory(Offset, InitCount: Integer);
begin
  FillChar(FItems^.Int8Array[Offset], InitCount, 0);
end;

const
  Quantum = 4;
{    :      
       ,   Quantum ; 
   TVector   Quantum = 4 }

function TVector.ItemsSize: Integer;
begin
{  Result:=(FCount + Quantum - 1) div Quantum * Quantum * FElemSize;}
  Result:=((FCount + 3) and -4) * FElemSize;
end;

procedure TVector.ChangeSize(ACount: Integer);
var
  NewCount, NewSize: Int32;
begin
  {$IFDEF CHECK_VECTORS}
  if ACount < 0 then Error(SRangeError, [ACount]);
  {$ENDIF}
{  NewCount:=(ACount + Quantum - 1) div Quantum * Quantum;}
  NewCount:=(ACount + 3) and -4;
  if NewCount <> FCapacity then begin
    NewSize:=NewCount * FElemSize;
    if NewSize > MaxBytes then Error(SDataTooLarge, [NewSize]);
    {$IFNDEF V_FREEPASCAL}
    {$IFDEF WIN32}
    ReAllocMem(FItems, NewSize);
    {$ELSE}
    FItems:=ReAllocMem(FItems, FCapacity * FElemSize, NewSize);
    {$ENDIF}
    {$ELSE}
    ReallocMem(FItems, NewSize);
    {$ENDIF}
    FCapacity:=NewCount;
  end;
  if NewCount > FCount then InitMemory(FCount * FElemSize, NewCount - FCount);
end;

procedure TVector.Resize(ACount: Integer);
{   ChangeSize ,     
    ,       
    MaxBytes;   Assign }
var
  NewCount: Int32;
begin
  NewCount:=(ACount + 3) and -4;
  if NewCount <> FCapacity then begin
    {$IFNDEF V_FREEPASCAL}
    {$IFDEF WIN32}
    ReAllocMem(FItems, NewCount * FElemSize);
    {$ELSE}
    FItems:=ReAllocMem(FItems, FCapacity * FElemSize, NewCount * FElemSize);
    {$ENDIF}
    {$ELSE}
    ReallocMem(FItems, NewCount * FElemSize);
    {$ENDIF}
    FCapacity:=NewCount;
  end;
end;

procedure TVector.SetCount(ACount: Integer);
begin
  ChangeSize(ACount);
  FCount:=ACount;
end;

procedure TVector.WriteToStream(VStream: TVStream);
begin
  VStream.WriteInt32(FCount);
  VStream.WriteProc(FElemSize, SizeOf(FElemSize));
  if FCount > 0 then VStream.WriteProc(FItems^, FCount * FElemSize);
end;

procedure TVector.ReadFromStream(VStream: TVStream);
begin
  Clear;
  FCount:=VStream.ReadInt32;
  VStream.ReadProc(FElemSize, SizeOf(FElemSize));
  if FCount > 0 then begin
    ChangeSize(FCount);
    VStream.ReadProc(FItems^, FCount * FElemSize);
  end;
end;

procedure TVector.RawAssign(Source: TVector);
begin
  if FElemSize = Source.FElemSize then begin
    Resize(Source.FCount);
    FCount:=Source.FCount;
    Move(Source.FItems^, FItems^, FCount * FElemSize);
  end
  else
    Error(SAssignError, [0]);
end;

procedure TVector.Assign(Source: TVector);
begin
  RawAssign(Source);
end;

procedure TVector.RawCopyRange(FromIndex: Integer; Source: TVector; SourceFrom,
  SourceTo: Integer);
var
  NewLength, CopyCount: Integer;
begin
  {$IFDEF CHECK_VECTORS}
  if (FromIndex < 0) or (SourceFrom < 0) or (SourceTo >= Source.Count) or
    (SourceFrom > SourceTo)
  then
    Error(SErrorInParameters, [0]);
  {$ENDIF}
  if FElemSize = Source.FElemSize then begin
    CopyCount:=SourceTo - SourceFrom + 1;
    NewLength:=FromIndex + CopyCount;
    if NewLength > Count then Count:=NewLength;
    Move((PChar(Source.FItems) + SourceFrom * FElemSize)^,
      (PChar(FItems) + FromIndex * FElemSize)^, CopyCount * FElemSize);
  end
  else
    Error(SAssignError, [0]);
end;

procedure TVector.CopyRange(FromIndex: Integer; Source: TVector; SourceFrom,
  SourceTo: Integer);
begin
  RawCopyRange(FromIndex, Source, SourceFrom, SourceTo);
end;

procedure TVector.RawExchangeRange(FromIndex, ToIndex, ACount: Integer);
begin
  {$IFDEF CHECK_VECTORS}
  if (FromIndex < 0) or (ToIndex < 0) or (FromIndex + ACount > Count) or
    (ToIndex + ACount > Count) or
    (IntMin(FromIndex, ToIndex) + ACount > IntMax(FromIndex, ToIndex))
  then
    Error(SErrorInParameters, [0]);
  {$ENDIF}
  MemExchange(
    (PChar(FItems) + FromIndex * FElemSize)^,
    (PChar(FItems) + ToIndex * FElemSize)^,
    ACount * FElemSize);
end;

procedure TVector.ExchangeRange(FromIndex, ToIndex, ACount: Integer);
begin
  RawExchangeRange(FromIndex, ToIndex, ACount);
end;

function TVector.RawEqualTo(V: TVector): Bool;
begin
  if FCount = V.FCount then
    Result:=MemEqual(FItems^, V.FItems^, FCount * FElemSize)
  else
    Result:=False;
end;

function TVector.EqualTo(V: TVector): Bool;
begin
  Result:=RawEqualTo(V);
end;

procedure TVector.Expand(I: Integer);
var
  Delta: Integer;
begin
  {$IFDEF CHECK_VECTORS}
  if (I < 0) or (I > FCount) then Error(SRangeError, [I]);
  {$ENDIF}
  if FCapacity <= FCount then begin
    if FCapacity > 64 then Delta:=UInt32(FCapacity) div 4
    else { UInt32(...): allows Delphi 5 to generate more efficient code }
      if FCapacity > 8 then Delta:=16
      else
        Delta:=4;
    ChangeSize(FCapacity + Delta);
  end;
  Delta:=I * FElemSize;
  Move(FItems^.Int8Array[Delta], FItems^.Int8Array[Delta + FElemSize],
    (FCount - I) * FElemSize);
  Inc(FCount);
end;

procedure TVector.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity < FCount then Error(SWrongVectorSize, [NewCapacity]);
  ChangeSize(NewCapacity);
end;

procedure TVector.Delete(I: Integer);
var
  Offset: Integer;
begin
  {$IFDEF CHECK_VECTORS}
  if (I < 0) or (I >= FCount) then Error(SRangeError, [I]);
  {$ENDIF}
  Offset:=I * FElemSize;
  Dec(FCount);
  Move(FItems^.Int8Array[Offset + FElemSize], FItems^.Int8Array[Offset],
    (FCount - I) * FElemSize);
  if FCapacity - FCount > Int32(UInt32(FCapacity) div 4) then
    ChangeSize(FCount);
end;

procedure TVector.DeleteRange(I, ACount: Integer);
var
  Offset: Integer;
begin
  if ACount > 0 then begin
    {$IFDEF CHECK_VECTORS}
    if (I < 0) or (I + ACount > FCount) then Error(SRangeError, [I]);
    {$ENDIF}
    Offset:=I * FElemSize;
    Dec(FCount, ACount);
    Move(FItems^.Int8Array[Offset + ACount * FElemSize], FItems^.Int8Array[Offset],
      (FCount - I) * FElemSize);
    ChangeSize(FCount);
  end;
end;

procedure TVector.Clear;
begin
  SetCount(0);
end;

function TVector.SizeInBytes: Integer;
begin
  Result:=SizeOf(FCount) + SizeOf(FElemSize) + SizeOf(FItems) +
    FCount * FElemSize;
end;

procedure TVector.Pack;
begin
  Capacity:=FCount;
end;

procedure TVector.Grow(Delta: Integer);
begin
  Count:=Count + Delta;
end;

{ TSortableVector }

procedure TSortableVector.DoSortRange(L, R: Integer);
var
  I, J: Integer;
  T: array [0..MaxElemSize - 1] of Byte; {!!!}
begin
  repeat
    I:=L;
    J:=R;
    GetUntyped((L + R) shr 1, T);
    repeat
      while Compare(I, T) < 0 do Inc(I);
      while Compare(J, T) > 0 do Dec(J);
      if I <= J then begin
        Exchange(I, J);
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then DoSortRange(L, J);
    L:=I;
  until I >= R;
end;

procedure TSortableVector.DoSortRangeDesc(L, R: Integer);
var
  I, J: Integer;
  T: array [0..MaxElemSize - 1] of Byte; {!!!}
begin
  repeat
    I:=L;
    J:=R;
    GetUntyped((L + R) shr 1, T);
    repeat
      while Compare(I, T) > 0 do Inc(I);
      while Compare(J, T) < 0 do Dec(J);
      if I <= J then begin
        Exchange(I, J);
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then DoSortRangeDesc(L, J);
    L:=I;
  until I >= R;
end;

procedure TSortableVector.SortRange(L, R: Integer);
begin
  {$IFDEF CHECK_VECTORS}
  if R >= Count then Error(SRangeError, [R]);
  {$ENDIF}
  if L < R then DoSortRange(L, R);
end;

procedure TSortableVector.SortRangeDesc(L, R: Integer);
begin
  {$IFDEF CHECK_VECTORS}
  if R >= Count then Error(SRangeError, [R]);
  {$ENDIF}
  if L < R then DoSortRangeDesc(L, R);
end;

procedure TSortableVector.Sort;
begin
  SortRange(0, Count - 1);
end;

procedure TSortableVector.SortDesc;
begin
  SortRangeDesc(0, Count - 1);
end;

procedure TSortableVector.DoSortRangeWith(L, R: Integer; AVector: TSortableVector);
var
  I, J: Integer;
  T: array [0..MaxElemSize - 1] of Byte; {!!!}
begin
  repeat
    I:=L;
    J:=R;
    GetUntyped((L + R) shr 1, T);
    repeat
      while Compare(I, T) < 0 do Inc(I);
      while Compare(J, T) > 0 do Dec(J);
      if I <= J then begin
        Exchange(I, J);
        AVector.Exchange(I, J); {    AVector }
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then DoSortRangeWith(L, J, AVector);
    L:=I;
  until I >= R;
end;

procedure TSortableVector.DoSortRangeDescWith(L, R: Integer; AVector: TSortableVector);
var
  I, J: Integer;
  T: array [0..MaxElemSize - 1] of Byte; {!!!}
begin
  repeat
    I:=L;
    J:=R;
    GetUntyped((L + R) shr 1, T);
    repeat
      while Compare(I, T) > 0 do Inc(I);
      while Compare(J, T) < 0 do Dec(J);
      if I <= J then begin
        Exchange(I, J);
        AVector.Exchange(I, J); {    AVector }
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then DoSortRangeDescWith(L, J, AVector);
    L:=I;
  until I >= R;
end;

procedure TSortableVector.SortRangeWith(L, R: Integer; AVector: TSortableVector);
begin
  {$IFDEF CHECK_VECTORS}
  if R >= Count then Error(SRangeError, [R]);
  {$ENDIF}
  if L < R then DoSortRangeWith(L, R, AVector);
end;

procedure TSortableVector.SortRangeDescWith(L, R: Integer; AVector: TSortableVector);
begin
  {$IFDEF CHECK_VECTORS}
  if R >= Count then Error(SRangeError, [R]);
  {$ENDIF}
  if L < R then DoSortRangeDescWith(L, R, AVector);
end;

procedure TSortableVector.SortWith(AVector: TSortableVector);
begin
  SortRangeWith(0, Count - 1, AVector);
end;

procedure TSortableVector.SortDescWith(AVector: TSortableVector);
begin
  SortRangeDescWith(0, Count - 1, AVector);
end;

procedure TSortableVector.SortWithArray(VectorArray: array of TSortableVector);
var
  N: Integer;

  procedure SpecialSortRange(L, R: Integer);
  var
    I, J, K: Integer;
    T: array [0..MaxElemSize - 1] of Byte; {!!!}
  begin
    repeat
      I:=L;
      J:=R;
      GetUntyped((L + R) shr 1, T);
      repeat
        while Compare(I, T) < 0 do Inc(I);
        while Compare(J, T) > 0 do Dec(J);
        if I <= J then begin
          Exchange(I, J);
          {    Vectors }
          for K:=0 to High(VectorArray) do
            VectorArray[K].Exchange(I, J);
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then SpecialSortRange(L, J);
      L:=I;
    until I >= R;
  end;

begin
  N:=Count - 1;
  if N > 0 then SpecialSortRange(0, N);
end;

procedure TSortableVector.SortDescWithArray(VectorArray: array of TSortableVector);
var
  N: Integer;

  procedure SpecialSortRangeDesc(L, R: Integer);
  var
    I, J, K: Integer;
    T: array [0..MaxElemSize - 1] of Byte; {!!!}
  begin
    repeat
      I:=L;
      J:=R;
      GetUntyped((L + R) shr 1, T);
      repeat
        while Compare(I, T) > 0 do Inc(I);
        while Compare(J, T) < 0 do Dec(J);
        if I <= J then begin
          Exchange(I, J);
          {    Vectors }
          for K:=0 to High(VectorArray) do
            VectorArray[K].Exchange(I, J);
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then SpecialSortRangeDesc(L, J);
      L:=I;
    until I >= R;
  end;

begin
  N:=Count - 1;
  if N > 0 then SpecialSortRangeDesc(0, N);
end;

procedure TSortableVector.Invert;
var
  I, N: Integer;
begin
  N:=Count;
  for I:=0 to UInt32(N) div 2 - 1 do begin
    Dec(N);
    Exchange(I, N);
  end;
end;

end.
